"""Pinecone API Client for the cheap developers."""
from __future__ import annotations

from dataclasses import dataclass, field
from os import environ
from typing import List

from aiofauna import APIClient
from aiohttp import ClientSession
from dotenv import load_dotenv

from ..schemas.pinecone import (Embedding, MetaData, Query, QueryBuilder,
                                QueryMatch, QueryRequest, QueryResponse,
                                UpsertRequest, UpsertResponse, Vector)

load_dotenv()


@dataclass(init=True, repr=True, unsafe_hash=False, frozen=False)
class PineconeClient(APIClient):
    """
    Cheapcone went greedy and removed the namespace feature from the free tier so let's query the API directly with MongoDB Filter Expressions.
    Also it's a good example of streamlining the development of API clients.
    """

    base_url: str = field(default=environ["PINECONE_API_URL"], init=True, repr=True)
    api_key: str = field(default=environ["PINECONE_API_KEY"], init=True, repr=False)

    def __load__(self) -> ClientSession:
        """Lazy load the client session."""
        return ClientSession(base_url=self.base_url, headers={"api-key": self.api_key})

    async def upsert(self, embeddings: List[Embedding]) -> UpsertResponse:
        """
        upsert
        Upsert embeddings into the vector index.

        Args:
            embeddings (List[Embedding]): Embeddings to upsert.

        Returns:
            UpsertResponse: Upsert response.
        """
        async with self.__load__() as session:
            values: List[Vector] = []
            metadata: List[MetaData] = []
            for embedding in embeddings:
                values.append(embedding.values)
                metadata.append(embedding.metadata)

            async with session.post(
                "/vectors/upsert",
                json={
                    "vectors": [
                        UpsertRequest(values=values, metadata=metadata).dict()
                        for values, metadata in zip(values, metadata)
                    ]
                },
            ) as response:
                return UpsertResponse(**await response.json())

    async def query(
        self, expr: Query, vector: Vector, includeMetadata: bool = True, topK: int = 4
    ) -> QueryResponse:
        """query
        Query the vector index.

        Args:
            expr (Query): Query expression.
            vector (Vector): Query vector.
            includeMetadata (bool, optional): Whether to include metadata in the response. Defaults to True.
            topK (int, optional): Number of results to return. Defaults to 10.

        Returns:
            QueryResponse: Query response.
        """
        async with self.__load__() as session:
            payload = QueryRequest(
                topK=topK,
                filter=expr,
                vector=vector,
                includeMetadata=includeMetadata,
            ).dict()
            async with session.post(
                "/query",
                json=payload,
            ) as response:
                return QueryResponse(**await response.json())



from __future__ import annotations

from uuid import uuid4

from aiofauna import Document
from pydantic import Field

from .typedefs import List, MetaData, Query, Value, Vector


class QueryBuilder:
    """Query builder for Pinecone Query API with MongoDB-like syntax."""

    def __init__(self, field: str = None, query: Query = None):  # type: ignore
        self.field = field
        self.query = query if query else {}

    def __repr__(self) -> str:
        return f"{self.query}"

    def __call__(self, field_name: str) -> QueryBuilder:
        return QueryBuilder(field_name)

    def __and__(self, other: QueryBuilder) -> QueryBuilder:
        return QueryBuilder(query={"$and": [self.query, other.query]})

    def __or__(self, other: QueryBuilder) -> QueryBuilder:
        return QueryBuilder(query={"$or": [self.query, other.query]})

    def __eq__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$eq": value}})

    def __ne__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$ne": value}})

    def __lt__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$lt": value}})

    def __le__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$lte": value}})

    def __gt__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$gt": value}})

    def __ge__(self, value: Value) -> QueryBuilder:
        return QueryBuilder(query={self.field: {"$gte": value}})

    def in_(self, values: List[Value]) -> QueryBuilder:
        """MongoDB-like syntax for $in operator."""
        return QueryBuilder(query={self.field: {"$in": values}})

    def nin_(self, values: List[Value]) -> QueryBuilder:
        """MongoDB-like syntax for $nin operator."""
        return QueryBuilder(query={self.field: {"$nin": values}})


class UpsertRequest(Document):
    """Request to upsert a vector."""

    id: str = Field(default_factory=lambda: str(uuid4()))
    values: Vector = Field(...)
    metadata: MetaData = Field(...)


class Embedding(Document):
    """Request to upsert a vector."""

    values: Vector = Field(...)
    metadata: MetaData = Field(...)


class QueryRequest(Document):
    """Request to query for similar vectors."""

    topK: int = Field(default=10)
    filter: dict = Field(...)
    includeMetadata: bool = Field(default=True)
    vector: Vector = Field(...)


class QueryMatch(Document):
    """A single match from a query."""

    id: str = Field(...)
    score: float = Field(...)
    metadata: MetaData = Field(...)


class QueryResponse(Document):
    """Response to a query."""

    matches: List[QueryMatch] = Field(...)


class UpsertResponse(Document):
    """Response to an upsert."""

    upsertedCount: int = Field(...)



"""Chat Completions Schemas"""
from typing import Dict, List, Literal

from aiofauna import Document, Field

Role = Literal["assistant", "user", "system", "function"]
Model = Literal["gpt-4-0613", "gpt-3.5-turbo-16k-0613"]
Size = Literal["256x256", "512x512", "1024x1024"]
Format = Literal["url", "base64"]

from typing import Dict, List, Literal, Union

from aiofauna import *
from aiofauna.typedefs import *

Vector = List[float]
Value = str | int | float | bool | List[str]
MetaData = Dict[str, Value]
Filter = Literal["$eq", "$ne", "$lt", "$lte", "$gt", "$gte", "$in", "$nin"]
AndOr = Literal["$and", "$or"]
Query = Union[
    Dict[str, Union[Value, "Query", List[Value], List["Query"]]],
    Dict[AndOr, List[Dict[str, Union[Value, "Query", List[Value], List["Query"]]]]],
]
Size = Literal["256x256", "512x512", "1024x1024"]
Format = Literal["url", "b64_json"]



from typing import *

from aiofauna import *

from .typedefs import *


class ChatMessage(Document):
    role: Role = Field(..., description="The role of the message")
    content: str = Field(..., description="The content of the message")


class ChatCompletionRequest(Document):
    model: Model = Field(..., description="The model to use for the completion")
    messages: List[ChatMessage] = Field(
        ..., description="The messages to use for the completion"
    )
    temperature: float = Field(
        default=0.2, description="The temperature of the completion"
    )
    max_tokens: int = Field(
        1024, description="The maximum number of tokens to generate"
    )
    stream: bool = Field(False, description="Whether to stream the completion or not")


class ChatCompletionUssage(Document):
    prompt_tokens: int = Field(..., description="The number of tokens in the prompt")
    completion_tokens: int = Field(
        ..., description="The number of tokens in the completion"
    )
    total_tokens: int = Field(..., description="The total number of tokens")


class ChatCompletionChoice(Document):
    index: int = Field(..., description="The index of the choice")
    message: ChatMessage = Field(..., description="The message of the choice")
    finish_reason: str = Field(..., description="The reason the choice was finished")


class ChatCompletionResponse(Document):
    id: str = Field(..., description="The id of the completion")
    object: str = Field(..., description="The object of the completion")
    created: int = Field(..., description="The creation time of the completion")
    model: Model = Field(..., description="The model used for the completion")
    choices: List[ChatCompletionChoice] = Field(
        ..., description="The choices of the completion"
    )
    usage: ChatCompletionUssage = Field(..., description="The usage of the completion")
    stream: bool = Field(..., description="Whether the completion was streamed or not")


class VectorResponse(Document):
    text: str = Field(..., description="The text of the completion")
    score: float = Field(..., description="The score of the completion")


class CreateImageResponse(Document):
    created: float = Field(...)
    data: List[Dict[Format, str]] = Field(...)


class CreateImageRequest(Document):
    """Creates an Image using Dall-E model from OpenAI.
    must use default values unless user prompts for a different configuration,
    will be use in case the user asks for an image that is not a logo or a photo.
    """

    prompt: str = Field(...)
    n: int = Field(default=1)
    size: Size = Field(default="1024x1024")
    response_format: Format = Field(default="url")